"""
Compatibility layer to run certbot both on Linux and Windows.

This package contains all logic that needs to be implemented specifically for Linux and for Windows.
Then the rest of certbot code relies on this module to be platform agnostic.
"""
